/****************************************************************************\
*  *
* File 'el_misc.cpp'  - miscellaneous routines for ElSocc                    *
* Project started     - 21:24:13, 19-01-98                                   *
* Last modified       - 21:24:27, 19-01-98                                   *
*                                                                            *
* Written by EXeLENZ (2:5030/448.1@fidonet, ivan@solvo.ru)                   *
*  *
\****************************************************************************/

#include <stdio.h>

#include "./egl_sdl/egl.h"
#include "./egl_sdl/egl_defs.h"

#include "fxffi.h"
#include "fxres.h"

/****************************************************************************\
* Function : max3                                                            *
* Desc     : Returns largest of 3 values                                     *
* Input    : 3 numbers                                                       *
* Output   : max                                                             *
\****************************************************************************/
int max3 (int a,
          int b,
          int c) {

   if (a>b && a>c) return a;
   if (b>a && b>c) return b;
   if (c>a && c>b) return c;

   return a;
}

inline int GetShift (unsigned long mask)
{
   int shift=0;
   for (; (mask & 1) == 0; mask >>= 1)
      shift++;

   return shift;
}

/****************************************************************************\
* Function : EGLloadFFI                                                      *
* Desc     : Loads BitMap freo resource                                      *
* Input    : handle                                                   *
* Output   : BitMap; NULL if failed                                          *
\****************************************************************************/
BitMap * EGLloadFFI (int handle) {

   unsigned char *pRes=ResLoad (handle);
   if (!pRes) return NULL;
   // the resource should be unloaded at program exit

   long ImgStart=* ((long *) (pRes+sizeof (FFI_FileHeader)));
   FFI_ImageHeader *pHdr=(FFI_ImageHeader *) (pRes+ImgStart);

   BitMap *pBitmap;
   EGLallocBlock (&pBitmap, pHdr->wXSize, pHdr->wYSize);
   pBitmap->Image = pRes+ImgStart+sizeof (FFI_ImageHeader);

   // perform pixel conversion, if needed
   if (pHdr->dwFlags & FFII_VALIDMASK) {
      ulong RMask, GMask, BMask, SrcRMask, SrcGMask, SrcBMask;

      EGLgetPixelFormat (&RMask, &GMask, &BMask);
      SrcRMask=pHdr->dwRMask;
      SrcGMask=pHdr->dwGMask;
      SrcBMask=pHdr->dwBMask;

      if (RMask != SrcRMask || GMask != SrcGMask || BMask != SrcBMask) {
         int SrcRShift=GetShift (SrcRMask);
         int SrcGShift=GetShift (SrcGMask);
         int SrcBShift=GetShift (SrcBMask);

         int SrcGBits=0;
         for (int SrcBit=0; SrcBit<32; SrcBit++)
            if ((SrcGMask & (1 << SrcBit)) != 0)
                SrcGBits++;

         /*EGLtrace ("SrcRShift %i, SrcGShift %i, SrcBShift %i, SrcGBits %i\n",
                   SrcRShift, SrcGShift, SrcBShift, SrcGBits);

         EGLtrace ("RMask %04X, GMask %04X, BMask %04X\n",
                   RMask, GMask, BMask);*/

         int DestRShift=GetShift (RMask);
         int DestGShift=GetShift (GMask);
         int DestBShift=GetShift (BMask);

         int DestGBits=0;
         for (int DestBit=0; DestBit<32; DestBit++)
            if ((GMask & (1 << DestBit)) != 0)
                DestGBits++;

         /*EGLtrace ("DestRShift %i, DestGShift %i, DestBShift %i, DestGBits %i\n",
                   DestRShift, DestGShift, DestBShift, DestGBits);*/

         ushort *p=(ushort *) pBitmap->Image;
         for (int i=0; i<pHdr->wXSize*pHdr->wYSize; i++, p++) {
            int R=(*p & SrcRMask) >> SrcRShift;

            int G=(*p & SrcGMask) >> SrcGShift;
            // 15/16 bpp
            if (SrcGBits > DestGBits)
               G >>= (SrcGBits - DestGBits);
            else if (SrcGBits < DestGBits)
               G <<= (DestGBits - SrcGBits);

            int B=(*p & SrcBMask) >> SrcBShift;

            *p = (ushort) (((R << DestRShift) & RMask) |
                           ((G << DestGShift) & GMask) |
                           ((B << DestBShift) & BMask));
         }
      }
   }

   return pBitmap;
}
