/* $Header: /C/FaerieX/FFIC/FFIC.cpp 1     20.07.98 22:35 Yole $
   FaerieFire Image Converter
   Written by Yoletir / FaerieFire AKA Dmitry Jemerov, 2:5030/654
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <fxFFI.h>

/////////////////////////////////////////////////////////////////////////////
// Types and structures

// Element of the list of images to be written to the .FFI
struct ImgListElem
{
  char *FName;
  ImgListElem *next;
};

// Structure used for reading input images
struct InputImage
{
  FFI_ImageHeader FFIHdr;
  int InBPP;
  DWORD dwRMask, dwGMask, dwBMask;
  BYTE *InPalette;            // NULL if input file has no palette
  BYTE *ImgData;              // raw image data
};

/////////////////////////////////////////////////////////////////////////////
// Global variables

// The list of images to be written, the current item
ImgListElem *ImgList = NULL, *ImgList_Cur = NULL;
int ImgList_Count=0, CurImage=0;

char OutFName [255] = "";               // Output file name
int OutBPP = 8;                         // Output bits per pixel
int Out_BytesPerPixel = 1;              // Output bytes per pixel
FILE *out = NULL;                       // Output file
int OutPalette = 0;                     // Include palette in output
int OutColorKey = 0;                    // Store color key in output file

FFI_FileHeader FileHdr;                 // Output file header
BYTE Palette [768];                     // Output file palette
long *OffsetMap;                        // Output file image offset map
DWORD ColorKey;

int MsgLevel = 1;                       // 0 - quiet, 1 - default, 2 - debug

/////////////////////////////////////////////////////////////////////////////
// General functions

//
// Abort()
// Exits to OS with error message

void Abort (char *Err, ...)
{
  // cleanup
  if (out)
  {
    fclose (out);
    unlink (OutFName);       // kill incomplete/invalid .FFI
  }
  fcloseall();
  
  char buf [256];
  va_list va;
  va_start (va, Err);
  vsprintf (buf, Err, va);
  va_end (va);
  puts (buf);
  exit (1);
}

////
///  fxHC()
//   Makes a high-color value from 3 components

inline WORD fxHC_16 (BYTE r, BYTE g, BYTE b)
{
  return ((WORD) (b & 0x1F) << 11) | ((WORD) (g & 0x3F) << 5) | (WORD) (r & 0x1F);
}

inline WORD fxHC_15 (BYTE r, BYTE g, BYTE b)
{
  return ((WORD) (b & 0x1F) << 10) | ((WORD) (g & 0x1F) << 5) | (WORD) (r & 0x1F);
}

DWORD DefaultRMask (int bpp)
{
  switch (bpp)
  {
  case 15: return 0x7C00;
  case 16: return 0xF800;
  case 24: return 0x0000FF;
  default: return 0;
  }
}

DWORD DefaultGMask (int bpp)
{
  switch (bpp)
  {
  case 15: return 0x3E0;
  case 16: return 0x7F0;
  case 24: return 0x00FF00;
  default: return 0;
  }
}

DWORD DefaultBMask (int bpp)
{
  switch (bpp)
  {
  case 15: return 0x001F;
  case 16: return 0x001F;
  case 24: return 0xFF0000;
  default: return 0;
  }
}

void GetDefaultMasks (int bpp, DWORD *RMask, DWORD *GMask, DWORD *BMask)
{
  *RMask = DefaultRMask (bpp);
  *GMask = DefaultGMask (bpp);
  *BMask = DefaultBMask (bpp);
}

BYTE ExtractMask (DWORD value, DWORD mask)
{
  DWORD SrcMask=0x80000000;
  int DestShift=7;
  BYTE result=0;

  for (int i=0; i<32; i++)
  {
    if ((mask & SrcMask) && DestShift >= 0)
    {
      if (value & SrcMask) result |= (1 << DestShift);
      DestShift--;
    }
    SrcMask >>= 1;
  }
  
  return result;
}

void MergeMask (DWORD *dest, DWORD value, DWORD mask)
{
  DWORD SrcMask=0x80, MaskMask=0x80000000;
  int DestShift=31;

  for (int i=0; i<32; i++)
  {
    if (mask & MaskMask)
    {
      if (value & SrcMask) *dest |= (1 << DestShift);
      SrcMask >>= 1;
    }
    DestShift--;
    MaskMask >>= 1;
  }
}

/////////////////////////////////////////////////////////////////////////////
// Image processing functions

////
///  Read_BMP()
//   Function for reading .BMP files

// FName is used only for error messages

void Read_BMP (FILE *f, char *FName, InputImage *ii)
{
  // Analyze the headers
  BITMAPFILEHEADER FileHdr;
  fread (&FileHdr, sizeof (BITMAPFILEHEADER), 1, f);
  if (FileHdr.bfType != 0x4D42)
    Abort ("[Err] .BMP file %s has invalid signature\n", FName);

  BITMAPINFOHEADER InfoHdr;
  fread (&InfoHdr, sizeof (BITMAPINFOHEADER), 1, f);
  if (MsgLevel >= 2)
    printf (".BMP file (%ix%i), %i bpp\n", 
      InfoHdr.biWidth, abs (InfoHdr.biHeight), InfoHdr.biBitCount);
  
  if (InfoHdr.biCompression != BI_RGB && InfoHdr.biCompression != BI_BITFIELDS)
    Abort ("[Err] .BMP file %s is compressed\n", FName);
  ii->InBPP = InfoHdr.biBitCount;
  
  ii->FFIHdr.wXSize = (WORD) InfoHdr.biWidth;
  BOOL bReverse=FALSE;
  if (InfoHdr.biHeight > 0)
  {
    bReverse = TRUE;
    ii->FFIHdr.wYSize = (WORD) InfoHdr.biHeight;
  }
  else ii->FFIHdr.wYSize = (WORD) -InfoHdr.biHeight;

  // Copy the palette
  int NumColors=InfoHdr.biClrUsed;
  if (!NumColors && InfoHdr.biBitCount == 8)
    NumColors=256;
  fseek (f, sizeof (BITMAPFILEHEADER) + InfoHdr.biSize, SEEK_SET);
  if (NumColors == 256)
  {
    ii->InPalette = new BYTE [768];
    BYTE *p=ii->InPalette;
    for (int i=0; i<256; i++)
    {
      RGBQUAD rgb;
      fread (&rgb, sizeof (RGBQUAD), 1, f);
      *p++ = rgb.rgbRed;
      *p++ = rgb.rgbGreen;
      *p++ = rgb.rgbBlue;
    }
  }
  else if (InfoHdr.biBitCount == 16)
  {
    if (InfoHdr.biCompression == BI_BITFIELDS)
    {
      fread (&ii->dwRMask, 4, 1, f);
      fread (&ii->dwGMask, 4, 1, f);
      fread (&ii->dwBMask, 4, 1, f);
    }
    else 
      GetDefaultMasks (15, &ii->dwRMask, &ii->dwGMask, &ii->dwBMask);
  }
  else if (InfoHdr.biBitCount == 24)
  {
    ii->dwBMask = 0x00FF0000;
    ii->dwGMask = 0x0000FF00;
    ii->dwRMask = 0x000000FF;
  }

  // Copy the bitmap
  int BytesPerLine = (InfoHdr.biBitCount / 8) * InfoHdr.biWidth;

  ii->ImgData = new BYTE [InfoHdr.biHeight * BytesPerLine];
  BYTE *dp;
  if (bReverse)
    dp=ii->ImgData + (InfoHdr.biHeight - 1) * BytesPerLine;
  else
    dp=ii->ImgData;
  
  for (int y=0; y<InfoHdr.biHeight; y++)
  {
    fread (dp, BytesPerLine, 1, f);
    if (bReverse)
      dp -= BytesPerLine;
    else
      dp += BytesPerLine;
  }
}

////
///  Read_RAW
//   Parse file dimensions string in pParams and read the raw file

void Read_RAW (FILE *f, char *FName, InputImage *ii, char *pParams)
{
  int RawWidth=0, RawHeight=0, RawBPP=0;
  
  // Parse .RAW parameters
  char *p1=pParams;
  char *p2=p1;
  while (*p2 && *p2 != ',') p2++;
  if (!*p2) Abort ("[Err] .RAW width not specified\n");
  *p2=0;
  RawWidth=atoi (p1);
  if (!RawWidth) Abort ("[Err] Invalid .RAW width %s\n", p1);
  
  p1=++p2;
  while (*p2 && *p2 != ',') p2++;
  if (!*p2) Abort ("[Err] .RAW height not specified\n");
  *p2=0;
  RawHeight=atoi (p1);
  if (!RawHeight) Abort ("[Err] Invalid .RAW height %s\n", p1);

  p1=++p2;
  if (!*p1) Abort ("[Err] .RAW bpp not specified\n");
  RawBPP=atoi (p1);
  if (!RawBPP) Abort ("[Err] Invalid .RAW bpp %s\n", p1);

  // Fill the InputImage structure
  ii->FFIHdr.wXSize = RawWidth;
  ii->FFIHdr.wYSize = RawHeight;
  ii->InBPP = RawBPP;

  // Read the .RAW data
  fseek (f, 0, SEEK_END);
  int FileLength=ftell (f);
  if (RawBPP == 15) RawBPP = 16;
  int RawLength=RawHeight * RawWidth * RawBPP / 8;
  if (MsgLevel >= 2)
    printf ("Expected file length %li\n", RawLength);
  if (RawLength != FileLength)
    printf ("[Warn] %s: .RAW file dimensions do not match size of file\n", FName);
  fseek (f, 0, SEEK_SET);
  ii->ImgData = new BYTE [RawLength];
  fread (ii->ImgData, (RawLength > FileLength) ? FileLength : RawLength, 1, f);

  GetDefaultMasks (ii->InBPP, &ii->dwRMask, &ii->dwGMask, &ii->dwBMask);
}

////
///  ConvertPalToRGB()
//   Converts an image with 8bpp palette to a 15/16/24 bpp image

void ConvertPalToRGB (BYTE *In, BYTE *Out, int OutBPP,
                      BYTE *Pal, int NPixels)
{
  if (MsgLevel >= 2)
    printf ("Converting 8 bit palette to %i bit RGB\n", OutBPP);
  BYTE *pIn = In;
  BYTE *pOut = Out;

  for (int i=0; i<NPixels; i++)
  {
    BYTE red=Pal [*pIn * 3];
    BYTE green=Pal [*pIn * 3 + 1];
    BYTE blue=Pal [*pIn * 3 + 2];
    if (OutBPP == 16)
    {
      WORD color=fxHC_16 (red >> 3, green >> 2, blue >> 3);
      * ((WORD *) pOut) = color;
      pOut += 2;
    }
    else if (OutBPP == 15)
    {
      WORD color=fxHC_15 (red >> 3, green >> 3, blue >> 3);
      * ((WORD *) pOut) = color;
      pOut += 2;
    }
    else if (OutBPP == 24)
    {
      *pOut++ = red;
      *pOut++ = green;
      *pOut++ = blue;
    }

    pIn++;
  }
}

//
// ConvertRGBToRGB()
// Converts an RGB image to another RGB image with different color depth

void ConvertRGBToRGB (const InputImage &ii, BYTE *Out, int OutBPP, int NPixels)
{
  BYTE *pIn = ii.ImgData;
  BYTE *pOut = Out;

  int InBytePP = (ii.InBPP == 15) ? 2 : ii.InBPP / 8;
  int OutBytePP = (OutBPP == 15) ? 2 : OutBPP / 8;

  DWORD OutRMask, OutGMask, OutBMask;
  GetDefaultMasks (OutBPP, &OutRMask, &OutGMask, &OutBMask);

  for (int i=0; i<NPixels; i++)
  {
    BYTE red, green, blue;
    DWORD hColor;

    memcpy (&hColor, pIn, InBytePP);
    pIn += InBytePP;

    red = ExtractMask (hColor, ii.dwRMask);
    green = ExtractMask (hColor, ii.dwGMask);
    blue = ExtractMask (hColor, ii.dwBMask);
    
    hColor = 0;
    MergeMask (&hColor, red, OutRMask);
    MergeMask (&hColor, green, OutGMask);
    MergeMask (&hColor, blue, OutBMask);
    
    memcpy (pOut, &hColor, OutBytePP);
    pOut += OutBytePP;
  }
}

//
// ProcessImage()
// Reads an input image file, performs all necessary conversions and
// writes the image to the output file

void ProcessImage (char *FName)
{
  static int GotPalette = 0;

  // Trim .RAW parameters specified in command line
  char *pcomma=strchr (FName, ',');
  if (pcomma)
    *pcomma++ = 0;

  printf ("Processing %s...\n", FName);

  FILE *f=fopen (FName, "rb");
  if (!f) Abort ("Cannot open image file %s\n", FName);

  InputImage ii;
  memset (&ii, 0, sizeof (InputImage));
  
  char Ext [4] = "";
  char *p=strrchr (FName, '.');
  if (p) 
  {
    strncpy (Ext, p+1, 3);
    Ext [3] = 0;
  }
  if (!stricmp (Ext, "BMP")) {
    if (pcomma) puts ("[Warn] File dimensions ignored\n");
    Read_BMP (f, FName, &ii);
  }
  else if (!stricmp (Ext, "RAW")) {
    if (!pcomma) Abort ("[Err] .RAW file dimensions not specified\n");
    Read_RAW (f, FName, &ii, pcomma);
  }
  else
    Abort ("[Err] Image file %s has unknown extension\n", FName);

  fclose (f);

  OffsetMap [CurImage] = ftell (out);

  if (ii.InBPP != 8 && ii.InBPP != 15 && ii.InBPP != 16 && ii.InBPP != 24)
    Abort ("[Err] File %s has invalid or unsupported BPP %i\n", FName, ii.InBPP);
  if (OutBPP == 8 && ii.InBPP > 8)
    Abort ("[Err] File %s: conversion of %i BPP to 8 BPP not supported\n", FName, ii.InBPP);
  
  ii.FFIHdr.dwPackedSize = ii.FFIHdr.wXSize * ii.FFIHdr.wYSize * Out_BytesPerPixel;
  if (OutColorKey)
  {
    ii.FFIHdr.dwFlags = FFII_COLORKEY;
    ii.FFIHdr.dwColorKey = ColorKey;
  }
  if (OutBPP == 15 || OutBPP == 16) {
    ii.FFIHdr.dwFlags |= FFII_VALIDMASK;
    if (OutBPP == 15) {
      ii.FFIHdr.dwBMask = 0x7C00;
      ii.FFIHdr.dwGMask = 0x03E0;
      ii.FFIHdr.dwRMask = 0x001F;
    } else {
      ii.FFIHdr.dwBMask = 0xF800;
      ii.FFIHdr.dwGMask = 0x07E0;
      ii.FFIHdr.dwRMask = 0x001F;
    }
  }

  fwrite (&ii.FFIHdr, sizeof (FFI_ImageHeader), 1, out);

  if (ii.InBPP == OutBPP)   // we need no conversion
    fwrite (ii.ImgData, ii.FFIHdr.dwPackedSize, 1, out);
  else {
    // convert the image to needed bpp
    BYTE *OutBuf = new BYTE [ii.FFIHdr.dwPackedSize];
    if (ii.InBPP == 8)
    {
      if (ii.InPalette != NULL)
        ConvertPalToRGB (ii.ImgData, OutBuf, OutBPP, ii.InPalette, 
          ii.FFIHdr.wXSize * ii.FFIHdr.wYSize);
      else if (GotPalette)
        ConvertPalToRGB (ii.ImgData, OutBuf, OutBPP, Palette, 
          ii.FFIHdr.wXSize * ii.FFIHdr.wYSize);
      else
        Abort ("[Err] Don't have the palette to convert image %s to %i bpp\n",
          FName, OutBPP);
    }
    else ConvertRGBToRGB (ii, OutBuf, OutBPP,
      ii.FFIHdr.wXSize * ii.FFIHdr.wYSize);

    fwrite (OutBuf, ii.FFIHdr.dwPackedSize, 1, out);
    delete OutBuf;
  }
  delete ii.ImgData;
        
  if (ii.InPalette != NULL)
  {
    if (!GotPalette)
    {
      GotPalette = 1;
      if (MsgLevel >= 2) printf ("Taking palette from file %s\n", FName);
      memcpy (Palette, ii.InPalette, 768);
    }
    delete ii.InPalette;
  }
}
 
/////////////////////////////////////////////////////////////////////////////
// Functions to process command line

//
// Help()
// Displays command line help

void Help()
{
  puts ("Usage:");
  puts ("  FFIC [options] filename[,w,h,bpp] [filename...]");
  puts (" or");
  puts ("  FFIC @filename");
  puts ("\nOptions:");
  puts ("  w,h,bpp      - width, height, and bits per pixel for .RAW files");
  puts ("  /O<filename> - set output file name");
  puts ("  /8           - make a 8 bpp output file (optionally with palette)");
  puts ("  /15          - make a high-color (15 bpp) output file");
  puts ("  /16          - make a high-color (16 bpp) output file");
  puts ("  /24          - make a true-color (24 bpp) output file");
  puts ("  /PAL         - include palette in output file (8 bpp only)");
  puts ("  /CK<hex clr> - specify the color key of output images\n");
  puts ("\nSupported input file formats: BMP, RAW");
  puts ("Input file formats are recognized by extension");

  exit (0);
}

//
// ProcessOption()
// Processes a single command line option

void ProcessOption (char *opt)
{
  *opt = toupper (*opt);
  if (*opt == '?' || *opt == 'H')
    Help();
  else if (*opt == 'O')
    strcpy (OutFName, opt+1);
  else if (*opt == '8')
  {
    OutBPP = 8;
    Out_BytesPerPixel = 1;
  }
  else if (!strcmp (opt, "15"))
  {
    OutBPP = 15;
    Out_BytesPerPixel = 2;
  }
  else if (!strcmp (opt, "16"))
  {
    OutBPP = 16;
    Out_BytesPerPixel = 2;
  }
  else if (!strcmp (opt, "24"))
  {
    OutBPP = 24;
    Out_BytesPerPixel = 3;
  }
  else if (!stricmp (opt, "PAL"))
    OutPalette = 1;
  else if (!stricmp (opt, "DEBUG"))
    MsgLevel = 2;
  else if (!strnicmp (opt, "CK", 2))
  {
    OutColorKey = 1;
    if (!sscanf (opt+2, "%x", &ColorKey))
    {
      printf ("[Err] Invalid color key option %s\n", opt);
      Help();
    }
  }
  else {
    printf ("[Err] Invalid command line option %s\n", opt);
    Help();
  }
}

//
// AddInputFile()
// Adds a file to the input file list. Doesn't check if it exists or anything else.

void AddInputFile (char *fname)
{
  if (!ImgList)
  {
    ImgList = new ImgListElem;
    // The default output file name is the name of the first input file with
    // extension changed to .FFI.
    if (!strlen (OutFName))
    {
      strcpy (OutFName, fname);
      char *p=strrchr (OutFName, '.');
      if (p) strcpy (p, ".ffi");
      else strcat (OutFName, ".ffi");
    }
    ImgList_Cur = ImgList;
  }
  else {
    ImgList_Cur->next = new ImgListElem;
    ImgList_Cur = ImgList_Cur->next;
  }

  ImgList_Cur->FName = new char [strlen (fname)+1];
  strcpy (ImgList_Cur->FName, fname);
  ImgList_Cur->next = NULL;
  ImgList_Count++;
}

//
// ProcessListFile()
// Processes a file containing options and input file names

void ProcessListFile (char *fname)
{
  FILE *f=fopen (fname, "rt");
  if (!f)
  {
    printf ("[Err] Cannot open options file %s\n", fname);
    return;  // it's not fatal
  }

  char arg [255];
  while (fscanf (f, "%s", arg) > 0)
  {
    if (arg [0] == '-' || arg [0] == '/')
      ProcessOption (arg+1);
    else
      AddInputFile (arg);
  }

  fclose (f);
}

/////////////////////////////////////////////////////////////////////////////
// main()

void main (int argc, char *argv[])
{
  puts ("FaerieFire Image Converter");
  puts ("Written by Yoletir / FaerieFire AKA Dmitry Jemerov, 2:5030/654");

  // Process command line
  //

  if (argc < 2) Help();
  for (int i=1; i<argc; i++)
  {
    if (argv [i][0] == '@')
      ProcessListFile (argv [i]+1);
    else if (argv [i][0] == '/' || argv [i][0] == '-')
      ProcessOption (argv [i]+1);
    else
      AddInputFile (argv [i]);
  }

  if (!ImgList) Abort ("[Err] Nothing to do. No input files specified.\n");
  if (OutPalette && OutBPP != 8)
  {
    puts ("[Warn] /PAL switch works only with 8 bpp output files");
    OutPalette = 0;
  }

  // Initialize output file
  //

  out=fopen (OutFName, "wb");
  if (!out) Abort ("[Err] Cannot open output file %s\n", OutFName);

  strncpy (FileHdr.Sign, "FaeI", 4);
  FileHdr.wImageCount = ImgList_Count;
  FileHdr.wBPP = OutBPP;
  FileHdr.dwFlags = 0;
  if (OutPalette) FileHdr.dwFlags |= FFIF_PALETTE;

  fwrite (&FileHdr, sizeof (FileHdr), 1, out);

  if (OutPalette)
  {
    memset (Palette, 0, sizeof (Palette));
    fwrite (Palette, sizeof (Palette), 1, out);
  }

  OffsetMap = new long [ImgList_Count];
  memset (OffsetMap, 0, ImgList_Count * sizeof (long));
  fwrite (OffsetMap, ImgList_Count * sizeof (long), 1, out);

  // Write the images
  //

  ImgList_Cur = ImgList;
  while (ImgList_Cur)
  {
    ProcessImage (ImgList_Cur->FName);
    ImgList_Cur = ImgList_Cur->next;
    CurImage++;
  }

  // Fixup the header
  //

  fseek (out, sizeof (FFI_FileHeader), SEEK_SET);
  if (OutPalette)
    fwrite (Palette, sizeof (Palette), 1, out);
  fwrite (OffsetMap, ImgList_Count * sizeof (long), 1, out);
  fclose (out);

  printf ("Successfully processed %i image(s)\n", ImgList_Count);
}   
