/* $Header: /C/FaerieX/FFRC/FFRC.cpp 1     20.07.98 22:36 Yole $
   FaerieFire resource compiler: main program file
   (C) 1998-2001 Dmitry Jemerov (2:5030/654, yole@spb.cityline.ru)
*/

#ifdef _WIN32
#include <io.h>
#else
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define __RES_HEADER
typedef int HRESELEM;
#include <fxres.h>

#include <zlib.h>

#ifdef WIN32
#define PROG_NAME "FFRCNT"
#else
#define PROG_NAME "FFRC"
#endif

// list of resource elements
struct ResElem
{
  ResDirEntry de;
  char *FileName;
  char *ResID;
  int indirect;
  ResElem *next;
};

ResHeader rh;

ResElem *reFirst=NULL, *reLast=NULL;
int reCount=0;
int CPPcomments=0;
int useZLib = 0;

FILE *lstf, *resf, *hf;

//
// AddResElem()
// adds elem to element list

void AddResElem (char *aFileName, char *aResID, int aIndirect)
{
  if (!reFirst) {
    reFirst=new ResElem;
    reLast=reFirst;
  } else {
    reLast->next = new ResElem;
    reLast=reLast->next;
  }
  reLast->next = NULL;
  reLast->FileName = new char [strlen (aFileName)+1];
  strcpy (reLast->FileName, aFileName);
  reLast->ResID = new char [strlen (aResID)+1];
  strcpy (reLast->ResID, aResID);
  reLast->indirect = aIndirect;
  reCount++;
}

//
// WriteComment()
// Writes a single comment line to header file

void WriteComment (char *s)
{
  if (CPPcomments)
    fprintf (hf, "// %s\n", s);
  else
    fprintf (hf, "/* %s */\n", s);
}

//
// help()
// Displays command line help

void help()
{
  puts ("\nUsage: "PROG_NAME" resname [options]");
  puts ("  resname.lst - resource list file");
  puts ("  resname.ffr - output resource file");
  puts ("  resname.h   - output header file\n");
  puts ("Options:");
  puts ("  /bind <filename> - bind resource to the end of filename");
  puts ("  /O <directory>   - specify resource output directory");
  puts ("  /cpp             - C++ style header file comments");
  puts ("  /zip             - compress resources with zlib\n");
  puts (PROG_NAME" /? - help");
  exit (1);
}

//
// LoadResourceList()
// Loads the <resname.lst> file and writes the header file

void LoadResourceList()
{
  printf  ("Loading resource list... ");
  char buf [256];
  char *p, *dp;
  char aFileName [80], aResID [80];
  int lstline=0, aIndirect=0;
  while (!feof (lstf))
  {
    lstline++;
    if (!fgets (buf, sizeof (buf)-1, lstf)) break;
    p=strrchr (buf, '\n');
    if (p) *p=NULL;
    p=buf;
    while ((*p == ' ') || (*p == '\t')) p++; //skip whitespace
    if (*p == ';')
    {
      p++;
      WriteComment (p);
      continue;
    }
    if (*p == 0)
    {
      fprintf (hf, "\n");
      continue;
    }
    if (*p == '!')
    {
      aIndirect = 1;
      p++;
    } else aIndirect=0;
    dp=aFileName;
    while ((*p != ' ') && (*p != '\t') && (*p != 0)) *dp++ = *p++; //TODO: check aFileName length
    *dp=0;
    while ((*p == ' ') || (*p == '\t')) p++; //skip whitespace
    if (*p == 0)
    {
      printf ("[Err] Resource list file, line %i: ResID not specified\n",
              lstline);
      fcloseall();
      exit (1);
    }
    dp=aResID;
    while ((*p != ' ') && (*p != '\t') && (*p != 0)) *dp++ = *p++; //TODO: check aResID length
    *dp = 0;
    AddResElem (aFileName, aResID, aIndirect);
    //write .H file entry
    if (reCount > 1)
      fprintf (hf, ",\n");
    fprintf (hf, "  %s = %i", aResID, reCount-1);
  }
}

static char *ResBuf=NULL, *ResPackBuf=NULL;
static long ResBufSize=0, ResPackBufSize=0;

//
// ReallocBuf()
// If the specified buffer is smaller than the needed size, grows it

void ReallocBuf (char **Buf, long *BufSize, int NeedSize)
{
  if (*BufSize < NeedSize)  //buffer too small, reallocate
  {
    if (*Buf) delete *Buf;
    *BufSize = NeedSize;
    *Buf = new char [NeedSize];
    if (!*Buf)
    {
      puts ("[Err] Cannot allocate resource buffer");
      fcloseall(); exit (1);
    }
  }
}

//
// WriteResourceElement()
// Writes a single element to the output file.

void WriteResourceElement (ResElem *reCur)
{
  FILE *ref=fopen (reCur->FileName, "rb");   // Resource Element File
  if (!ref)
  {
    printf ("[Err] Cannot open resource element file %s\n", reCur->FileName);
    fcloseall();
    exit (1);
  }

  //set up ResDirEntry struct
  fseek (ref, 0, SEEK_END);
  long ResLen=ftell (ref);
  fseek (ref, 0, SEEK_SET);
  reCur->de.Offset = ftell (resf);
  reCur->de.__padding = 0;
  if (reCur->indirect)
  {
    reCur->de.RealSize = 0;
    // convert FileName to full path
    char buf [256];
    _fullpath (buf, reCur->FileName, sizeof (buf)-1);
    delete reCur->FileName;
    reCur->FileName = new char [strlen (buf)+1];
    strcpy (reCur->FileName, buf);
    reCur->de.PackedSize = strlen (reCur->FileName)+1;
    reCur->de.Type = 0xFF00;
    fwrite (reCur->FileName, strlen (reCur->FileName)+1, 1, resf);
  }
  else {
    reCur->de.RealSize = ResLen;
    reCur->de.Type = 0;

    //append ref to resf
    ReallocBuf (&ResBuf, &ResBufSize, ResLen);
    fread (ResBuf, ResLen, 1, ref);
    if (useZLib)
    {
      // leave space for zlib
      ReallocBuf (&ResPackBuf, &ResPackBufSize, ResLen+10000);
      unsigned long packSize = ResPackBufSize;
      compress ((unsigned char *) ResPackBuf, &packSize,
        (unsigned char *) ResBuf, ResLen);
      reCur->de.PackedSize = packSize;
      fwrite (ResPackBuf, packSize, 1, resf);
    }
    else {
      reCur->de.RealSize = ResLen;
      fwrite (ResBuf, ResLen, 1, resf);
    }
  }
  fclose (ref);
}

/////////////////////////////////////////////////////////////////////////////
// main()

int main (int argc, char *argv[])
{
  puts ("FaerieFire Resource Compiler");
  puts ("Written by Yoletir / FaerieFire AKA Dmitry Jemerov, 2:5030/654");

  if (argc < 2) help();
  if ((!strcmp (argv [1], "/?")) || (!strcmp (argv [1], "-?"))) help();

  // prepare filename
  char fname [255], fullname [255], adrive [_MAX_DRIVE], adir [_MAX_DIR];
  char in_dir [255], out_dir [255];

  _splitpath (argv [1], adrive, adir, fname, NULL);
  strcpy (in_dir, adrive);
  strcat (in_dir, adir);
  if (!strlen (in_dir))
  {
    _splitpath (argv [0], adrive, adir, NULL, NULL);
    strcpy (in_dir, adrive);
    strcat (in_dir, adir);
  }
  strcpy (out_dir, in_dir);

  // parse other options
  int bind=0;
  char bindfname [255];
  long BindPos;
  for (int arg=2; arg<argc; arg++)
  {
    if (argv [arg][0] == '-') argv [arg][0] = '/';
    if (!strcmpi (argv [arg], "/BIND"))
    {
      if (arg==argc-1)
      {
        puts ("[Err] /BIND filename not specified");
        exit (1);
      }
      bind=1;
      strcpy (bindfname, argv [++arg]);
    }
    else if (!strcmpi (argv [arg], "/O"))
    {
      if (arg == argc-1)
      {
        puts ("[Err] /O directory not specified");
        exit (1);
      }
      strcpy (out_dir, argv [++arg]);
      if (out_dir [strlen (out_dir) - 1] != '\\')
        strcat (out_dir, "\\");
    }
    else if (!strcmpi (argv [arg], "/CPP"))
      CPPcomments=1;
    else if (!strcmpi (argv [arg], "/ZIP"))
      useZLib = 1;
    else {
      printf ("[Err] Unknown parameter %s\n", argv [arg]);
      help();
    }
  }

  // open files
  sprintf (fullname, "%s%s.lst", in_dir, fname);
  lstf=fopen (fullname, "rt");
  if (!lstf)
  {
    printf ("[Err] Cannot open resource list file %s\n", fullname);
    exit (1);
  }

  if (!bind)
  {
    sprintf (fullname, "%s%s.ffr", out_dir, fname);
    resf=fopen (fullname, "wb"); //TODO: ask for overwrite
    if (!resf)
    {
      printf ("[Err] Cannot open output resource file %s\n", fullname);
      exit (1);
    }
    BindPos = 0;
  }
  else {
    if (access (bindfname, 0))
      puts ("[Warn] Bind file doesn't exist");
    resf=fopen (bindfname, "r+b");
    if (!resf)
    {
      printf ("[Err] Cannot open bind file %s\n", bindfname);
      exit (1);
    }
    fseek (resf, 0, SEEK_END);
    BindPos = ftell (resf);
  }

  sprintf (fullname, "%s%s.h", out_dir, fname);
  hf=fopen (fullname, "wt");
  if (!hf)
  {
    printf ("[Err] Cannot open output header file %s\n", fullname);
    exit (1);
  }

  //write header header :)
  char buf [128];
  fprintf (hf, "#ifndef __RES_HEADER\n#define __RES_HEADER\n");
  sprintf (buf, "Resource header for %s.ffr", fname);
  WriteComment (buf);
  WriteComment ("Generated by FaerieFire resource compiler");
  fprintf (hf, "\n");
  fprintf (hf, "typedef enum {\n");

  //read and parse resource list file
  LoadResourceList();
  if (reCount == 0)
  {
    puts ("[Err] No resources specified in list file");
    fcloseall();
    exit (1);
  }
  fprintf (hf, "\n};\n\n#endif  /* __RES_HEADER */\n");
  printf ("%i resources.\n", reCount);

  //write resource header
  if (!bind)
    puts ("Writing resource file...");
  else
    puts ("Writing bound resource file...");

  if (!useZLib)
    strncpy (rh.Sign, "FaeR", 4);
  else
    strncpy (rh.Sign, "FaeZ", 4);

  rh.DirOffset=0;
  rh.ResCount=reCount;
  fwrite (&rh, sizeof (rh), 1, resf);

  //write resource elements
  for (ResElem *reCur=reFirst; reCur; reCur=reCur->next)
    WriteResourceElement (reCur);

  //fixup and write header
  rh.DirOffset = ftell (resf);
  fseek (resf, BindPos, SEEK_SET);
  fwrite (&rh, sizeof (rh), 1, resf);
  fseek (resf, 0, SEEK_END);

  //write directory
  for (reCur=reFirst; reCur; reCur=reCur->next)
    fwrite (&reCur->de, sizeof (ResDirEntry), 1, resf);

  if (bind)
    fwrite (&BindPos, 4, 1, resf); //write bound position

  //terminate
  fcloseall();
  delete ResBuf;
  delete ResPackBuf;
  return 0;
}
